#ifndef DBTOOLS_H_
#define DBTOOLS_H_

/**
 * \file dbtools.h
 * \brief Header file for different modules. Generally covers
 * database routines.
 * 
 * This header contains declarations of functions, structures
 * and constants related to TINE database file access.
 * This includes csv file access as well as address resolution
 * and caching routines.
 * 
 * Functions found here are implemented in
 * <ol>
 * <li>dblib.c</li>
 * <li>srvdbase.c</li>
 * <li>others (not yet completely analyzed)</li>
 * </ol>
 * 
 * \todo list all modules for which this is a header.
 */

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define EQPDB_FILE_TEMP "eqpdbase.tmp"
#define FECDB_FILE_TEMP "fecaddr.tmp"

#define FECXML_FILE "fec.xml"

#ifndef MAX_NUM_USERS
# define MAX_NUM_USERS 100
#endif

#ifndef MAX_NUM_IPNETS
# define MAX_NUM_IPNETS 100
#endif

/**
 * \internal
 * \brief A structure which contains information about a column in a CSV file.
 * 
 * This structure is used internally by CSVDBStruct.
 */
typedef struct
{
  char *nam;          /**< column name */
  char *def;          /**< default value */
  int  pos;           /**< column index position (to be determined) */
  int  fmt;           /**< value format */
  int  siz;           /**< number of element of specified format */
  BYTE *ptr;          /**< pointer to container position (initially relative pointer) */
  int (*fcn)(char *); /**< optional dispatch function (called when value has been determined) */
  BYTE *start;        /**< pointer starting position in container  (initially 0) */
} CSVCOLUMN;

/* DOS will #define this to use a 'far' segment */
#ifndef _CSVCOLUMN
# define _CSVCOLUMN CSVCOLUMN
#endif

#define CSVDB_PREALLOC 0x01

/**
 * \internal
 * \brief Container for the contents of a CSV file.
 * 
 * This structure has two purposes. It serves as
 * <ul>
 * <li>a container</li>
 * <li>its own single iterator</li>
 * </ul>
 * 
 * \see CSVCOLUMN
 * \todo redevelopment of csv library desirable
 */
typedef struct CSVDBStruct
{
  _CSVCOLUMN *col;                   /**< pointer to CSVCOLUMN list */
  int ncols;                         /**< num columns */
  int siz;                           /**< size of db file excluding comments => num rows - header */
  void (*ini)(struct CSVDBStruct *); /**< optional initializer routine */
  int  (*itr)(struct CSVDBStruct *); /**< optional iterator */
  void (*exi)(struct CSVDBStruct *); /**< optional finalizer routine */
  int cnt;                           /**< counter */
  BYTE *tgt;                         /**< pointer to the container */
  int tgt_siz;                       /**< size of container (e.g. sizeof(some typedef) */
  int tgt_flgs;                      /**< container flags (e.g. CSVDB_PREALLOC) */
} CSVDB;

#define USERS_ALLOWED_SETNAME "users_allowed"
#define USERS_DENIED_SETNAME "users_denied"
#define NETS_ALLOWED_SETNAME "nets_allowed"
#define NETS_DENIED_SETNAME "nets_denied"
#define NETS_NETWORK_SETNAME "nets_network"

typedef struct
{
	int (* processOpeningTag)( char * tag );
	int (* processClosingTag)( char * tag );
	int (* processData)( char * data );
	void * currentDataSection;
	int currentDataType;
  int currentDataCapacity;
	void * currentData;
} XML_DBHANDLER;

typedef struct listItem
{
  struct listItem * previous;
  struct listItem * next;
  void * data;
  int dataBufferAllocated;
} DBLNKLIST_ITEM;

typedef struct
{
  int nElements;
  DBLNKLIST_ITEM * first;
  DBLNKLIST_ITEM * current;
  DBLNKLIST_ITEM * last;
} DBLNKLIST;

TINE_EXPORT DBLNKLIST *XMLListCreate( void );
TINE_EXPORT void XMLListDelete( DBLNKLIST * list );
TINE_EXPORT void XMLListAdd( DBLNKLIST * list, void * data );
TINE_EXPORT void XMLListAddCopy( DBLNKLIST * list, void * data, int sizeOfData );
TINE_EXPORT void XMLListGoToFirst( DBLNKLIST * list );
TINE_EXPORT void XMLListGoToLast( DBLNKLIST * list );
TINE_EXPORT void XMLListRemoveCurrent( DBLNKLIST * list );
TINE_EXPORT void *XMLListGetNext( DBLNKLIST * list );
TINE_EXPORT void *XMLListGetPrevious( DBLNKLIST * list );
TINE_EXPORT int XMLListHasNext( DBLNKLIST * list );
TINE_EXPORT int XMLListHasPrevious( DBLNKLIST * list );
TINE_EXPORT void XMLSetCurrentHandler( XML_DBHANDLER * xmlHandler);
TINE_EXPORT void XMLRestoreHandler(void);
int dbpush( void * p );
int dbpop( void ** p );

typedef struct
{
  char name[SET_NAME_SIZE];
  DBLNKLIST *members;
} xmlSETIS;

typedef struct
{
  char almTag[32];
  char dataFormat[64];
  char almText[64];
  char devText[64];
  char datText[64];
  char urlText[64];
  UINT32 code;
  UINT32 mask;
  UINT32 severity;
  UINT32 system;
  UINT32 dataSize;
} xmlADS;

typedef struct
{ 
  UINT32 hstIndex;
  char hstDevName[DEVICE_NAME_SIZE];
  UINT32 hstHeartbeat;
  UINT32 hstPollRate;
  UINT32 hstArchRate;
  char hstTolerance[64];
  UINT32 hstDepthShort;
  UINT32 hstDepthLong;
  char hstRemoteServer[PROPERTY_REDIR_SIZE];
} xmlHIS;

typedef struct
{ 
  char almDevName[DEVICE_NAME_SIZE];
  char almValNormal[64];
  UINT32 almSevHigh;
  UINT32 almSevLow;
  UINT32 almSevHighWarn;
  UINT32 almSevLowWarn;
  UINT32 almSystem;
  UINT32 almValMask;
  UINT32 almCntThreshold;
  float  almValHigh;
  float  almValLow;
  float  almValHighWarn;
  float  almValLowWarn;
  int  code;              
  int  codeHigh;          
  int  codeLow;           
  char tagHigh[ALARM_TAG_SIZE]; 
  char tagLow[ALARM_TAG_SIZE]; 
} xmlAWIS;

typedef struct
{ 
  char name[PROPERTY_NAME_SIZE];
  char description[128];
  char redirection[PROPERTY_REDIR_SIZE];
  char deviceSet[SET_NAME_SIZE];
  char access[ACCESS_STR_SIZE];
  char dataInType[64];
  char dataOutType[64];
  char xEgu[16];
  char yEgu[16];
  UINT32 dataInSize;
  UINT32 dataOutSize;
  UINT32 id;
  float  xMax;
  float  xMin;
  float  yMax;
  float  yMin;
  DBLNKLIST *histories;
  DBLNKLIST *almwatches;
} xmlPRPIS;

typedef struct
{ 
  char name[DEVICE_NAME_SIZE];
  char desc[DEVICE_DESC_SIZE];
  char redirection[DEVICE_REDIR_SIZE];
  char propertySet[SET_NAME_SIZE];
  UINT32  number;
} xmlDEVIS;

typedef struct
{ 
  char name[DEVICE_NAME_SIZE];
  char target[DEVICE_NAME_SIZE];
} xmlAIS;

typedef struct
{ 
  char name[EQM_NAME_SIZE];
  char context[CONTEXT_NAME_SIZE];
  char serverName[EXPORT_NAME_SIZE];
  char serverGroup[EXPORT_NAME_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
  char serverMaster[EXPORT_NAME_SIZE];
  char slaveMaster[EXPORT_NAME_SIZE];
  UINT32  serverGroupIndex;
  UINT32  deviceSpace;
  DBLNKLIST *devices;
  DBLNKLIST *properties;
  DBLNKLIST *sets;
  DBLNKLIST *alarmdefs;
} xmlXIS;

typedef struct
{ 
  char name[FEC_NAME_SIZE];
  char context[CONTEXT_NAME_SIZE];
  char description[FEC_DESC_SIZE];
  char hardware[FEC_HDW_SIZE];
  char location[FEC_LOCATION_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
  char historyHome[HISTORY_HOME_SIZE];
  int  portOffset;
  DBLNKLIST *eqms;
  DBLNKLIST *sets;
  DBLNKLIST *alias;
} xmlFIDS;

extern XML_DBHANDLER DocumentHandler,FecHandler,EqmHandler,DeviceHandler;
extern XML_DBHANDLER PropertyHandler,AlmDefHandler,NameSetHandler,MemberHandler,AliasHandler,AlmWatchHandler,HistoryHandler;
TINE_EXPORT int XMLReadFile(char *filepath, char *filename, DBLNKLIST *fecList);
extern DBLNKLIST *gFecInfoList;
extern int isUsingXMLFecDatabase;

#if !defined(UNIX) && !defined(VXWORKS)
TINE_EXPORT char *strtok_r(char *s, const char *delim, char **ptrptr);
#endif
TINE_EXPORT char *strtrm(char *s);
TINE_EXPORT char *strntrm(char *s,int maxlen);
TINE_EXPORT int csvAssignValue(void *ptr,int fmt,int siz,char *str);
TINE_EXPORT int csvReadFile(char *path,char *file,CSVDB *db,void **target);
TINE_EXPORT int csvWriteFile(char *path,char *file,CSVDB *db);
TINE_EXPORT void csvStdInit(struct CSVDBStruct *db);
TINE_EXPORT int  csvStdIterator(struct CSVDBStruct *db);
TINE_EXPORT void csvStdExit(struct CSVDBStruct *db);
TINE_EXPORT int GetCsvColumnIndex(const char *csvHeader,const char *columnTag);
TINE_EXPORT char *GetCsvColumnString(int col,const char *str);
int GetDataBaseSize(char *filename);
int getFecIndex(char *fecName);
int locateExport(char *expName,char *context,int *rpcidx);
int asyncLocateExport(char *expName,char *context);
int GetAddressFromNameFile(char *filepath,char *context,char *expName, char *fecName,int *rpcidx);
int GetAddressFromNameServer(char *context,char *expName, char *fecName,int *rpcidx);
int addNameServerToTables(void);
int appendAddressToCache(char *fecName,char *ip, short portOffset);
int appendServerToCache(char *fecName,char *context,char *expName, char *eqmName);
int namcmp(char *name,char *str,int col,int limit);
int findcol(char *str,char *tag);
int getColumnStr(int col,char *str,char *buf,int bufsiz);
char *colptr(int col,char *str);
void BytesToHex(BYTE *b,char *s,int n);
void HexToBytes(char *s,BYTE *b,int nibs);
unsigned int HexToDec(char *str);
int GetNameServerAddress(void);
int getFecName(void);
int reserveFecAddressSpace(void);
int rmvRegUserLst(char *eqm,NAME16 *userlist,int listsize);
int getRegisteredUsers(NAME16 **ul,char *fn,int *nr,char *accstr,int *accflg,char *label);
int getAllowedUsers(char *eqm);
int getAllowedNets(const char *eqm);
void dumpRegisteredUsers(void);
int getRegisteredExports(char *filename,char *eqm);
int outputConsumerList(void);
int outputContractList(void);
int outputEqmList(void);
int outputEqmProperties(char *eqm);
int outputEqmDevices(char *eqm);
int outputContractStats(int i);
int outputConnectionStats(int k);
int outputServerStats(void);
int outputServerTime(void);
int outputSettings(void);
int dumpSettings(void);
int dumpContractStats(int i);
int dumpConnectionStats(int i);
int dumpConsumerStats(int i);
int dumpServerStats(void);
int dumpServerTime(void);
int dumpVersion(void);
TINE_EXPORT void isEquipmentNameServer(int value);
TINE_EXPORT int GetHostFromFecName(char *fecName,char *hostname);
TINE_EXPORT void FlushFecAddrCache(void);
TINE_EXPORT void flushFecAddrCache(void);
TINE_EXPORT void flushRedirectedLinkList(void);
TINE_EXPORT int GetENSAddr(char *name,char *addr);
TINE_EXPORT void ToggleNameServer(void);
int makeDirectoryPath(char *dname);

extern TINE_EXPORT int gIgnoreVolatileAddressCache;
extern TINE_EXPORT int gIgnoreAddressCacheOnUnknownAddress;
extern int gIsEquipmentNameServer;
extern FILE *dbgfp;
extern char dbgbuf[256];
extern char tagNameFilter[64];
extern int dbaseloaded;
extern int MaxTblAllocationSize;
extern char tineHomePath[80];
extern int NrOfDeniedUsers;
extern char fecDbPath[80];
extern int FecNameRegistered;
extern int ExportsRegistered;
extern int gotNameServerAddress;
extern NAME16 *gAllowedUsersList;
extern NAME16 *gDeniedUsersList;
extern int isNameServerRequest;
extern int NameServerLoaded;
extern int MaxNumAllowedUsers;
extern int MaxNumAllowedIPNets;
int addStandAloneEntryToAddressCache(char *ctxName,char *expName,char *eqmName,int *fecIdx,int *srvIdx);
#ifndef FS_RDONLY
extern char gAddrCacheFilePath[128];
int isAddressInCacheFile(char *filepath,int srvidx,int fecidx);
int addAddressToCacheFile(char *filepath,int srvidx,int fecidx);
void establishAddrCacheFilePath(void);
#endif
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* DBTOOLS_H_ */
