#ifndef GLOBALS_H_
#define GLOBALS_H_

#include "project.def"
#include "srvcore.h"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef CAN_SENDGLOBALS
# define CAN_SENDGLOBALS TRUE /* default behavior for opening the globals socket */
#endif
#ifndef GLOBALGRACE
# define GLOBALGRACE  3      /* retries for missing heartbeat */
#endif

#define GCAST_PROTOCOL_LEVEL 4
/**
 * \internal
 * \brief Legacy Structure used to managed global data.
 */
#define KEYWORD_NAME_SHORTSIZE 16
#define KEYWORD_NAME_SIZE 64
typedef struct         /* net */
{
  char  name[KEYWORD_NAME_SHORTSIZE]; /* global element name  (16 bytes) */
  short size;          /* number of elements gcast under "name" (2 bytes) */
  short format;        /* TINE data format  (2 bytes) */
  SINT32 time;         /* timestamp of gcast data (4 bytes) */
} GlbDataHdr3;
#define GLOBALDATA_HEADER3_SIZE (KEYWORD_NAME_SHORTSIZE + 2 + 2 + 4) /* 24 bytes */

/**
 * \internal
 * \brief Legacy Structure used to managed global data.
 */
typedef struct         /* net */
{
  char name[KEYWORD_NAME_SIZE]; /* global element name  (64 bytes) */
  short size;          /* number of elements gcast under "name" (2 bytes) */
  short format;        /* TINE data format  (2 bytes) */
  short status;
  short reserved;
  int dStamp;
  int sysStamp;
  SINT32 dTimeStamp;   /* timestamp of gcast data (4 bytes) */
  SINT32 dTimeStampMSEC;  
} GlbDataHdr;
#define GLOBALDATA_HEADER_SIZE (KEYWORD_NAME_SIZE + 2*4 + 4*4) /* 88 bytes */

typedef struct
{
  char context[CONTEXT_NAME_SIZE];/**< global context */
  char server[EXPORT_NAME_SIZE]; /**< global server : usually GLOBSRV */
  char keyword[KEYWORD_NAME_SIZE];/**< global keyword */
  short size;                /**< number of elements registered in global array */
  short format;              /**< globals data format */
  time_t time;               /**< timestamp of globals data */
  SINT32 timeMSEC;           /**< timestamp of globals data (msec) */
  struct timeval tv;         /**< timestamp of globals update */
  BYTE *data;                /**< data pointer */
  BYTE *refdata;             /**< for comparison */
  int  minPeriod;            /**< minimum period between multicasts */
  int  maxPeriod;            /**< maximum period between multicasts */
  double tolerance;          /**< allowed deviation of data (data units) */
  UINT32 srvIp;              /**< the originating server ip addr */
  UINT32 mcastGrp;           /**< the multicast group ip addr */
  UINT32 cbId;               /**< private callback ID */
  TCBFCNP cbfcn;             /**< callback function */
  short nlnks;               /**< number of common links */
  short isPending;           /**< pending flag */
  short isGrouped;           /**< grouped flag */
  short isBeingSent;         /**< busy flag */
  short disabled;            /**< disabled flag */
  short hasUserCallbackId;   /**< flag : TRUE if set by caller, otherwise false */
  short lastsize;            /**< number of elements in last incoming/outgoing packet */
  short status;              /**< current status */
  int conTblId;              /**< xref into conTbl if relevant */
  int dataStamp;             /**< current data stamp */
  int sysStamp;              /**< current system stamp */
  short hasReference;        /**< TRUE => use cb prototypes with reference */
  short isInsideCallback;    /**< inside callback flag */
  int grpLatency;            /**< notification lag of this group element (if grouped) */
  int syncOffset;            /**< learned offset in sysStamp of this group element (if grouped) */
  GrpTblEntry *grp;          /**< pointer to the group table entry (if relevant) */
  void *cbRef;               /**< reference pointer passed in recvNetGlobal */
} GlobalListStruct;

#ifndef GLOBALSRV
# define GLOBALSRV "GLOBALS"
#endif
#ifndef GLBTBL_CAPACITY
# define GLBTBL_CAPACITY 100
#endif
#ifndef GLOBALHEARTBEAT
# define GLOBALHEARTBEAT 1
#endif

extern int nglobals, ngcasts;
extern int GlobalHeartbeat;
extern int MaxNumGlobals;
extern int MaxNumGCasts;
extern int GlbTblCapacity;
extern int GCastTblCapacity;

extern GlobalListStruct **gcastTbl;   /* the server's send list */
extern GlobalListStruct **glbTbl; /* the client's recv list */

TINE_EXPORT GlobalListStruct *GetGlobalListElement(int id);
/**
 * \internal
 */
int parseGlobalsKeyword(char *key,char **context,char **server,char **keyword);
/**
 * \internal
 */
int getGlobalTableIndex(int start,char *context,char *server,char *keyword);

/* sendNetGlobal() */
/**
 * \brief registers and sends the accompanying keyword and data as a network global.
 *
 * The keyword supplied is appended to the globals table if not already registered and sent
 * immediately as a netork global per multicast if multicasting is enabled and/or per broadcast
 * if a broadcast nets table has been registered (via the inclusion of a ipbcast.csv or ipxbcast.csv
 * file in the startup database set).  Multicasts/Broadcasts will then be scheduled according the the minPeriod,
 * maxPeriod, and tolerance specified, in which case the din data object should contain a reference 
 * to a data object which will be regularly updated by the server code.  Alternatively, the 
 * call can specify '-1' as minPeriod and maxPeriod, in which case the network global will only be sent
 * upon a call to sendNetGlobal().
 * If a callback function is supplied, is it called following the successful transmission of the
 * network global.
 * 
 * Returns 0 upon successful globals transmission, otherwise a TINE error code.
 *
 * @param keyword is the globals KEYWORD which identifies the data set being sent.
 * @param din is a pointer to the input data set, that is, the data set to be sent from the server 
 * If din is a NULL pointer or contains a NULL data set, sendNetGlobal() returns an error.
 * @param callback is the address of the user-supplied callback routine to be called when data are sent 
 * from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * @param minPeriod is the minimum period in milliseconds between scheduled multi/broadcasts
 * @param maxPeriod is the maximum period in milliseconds between scheduled multi/broadcasts.  If the data are
 * still within tolerance at the end of this period, the data set will nonetheless be sent as a network
 * global.
 * @param tolerance is the allowed range between the current and previous data sets.  If exceeded, the
 * current data set is then sent as a network global.
 *  
 */
TINE_EXPORT int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),
                  int minPeriod,int maxPeriod,double tolerance);

/* AssignDataStampsToGlobal() */
/**
 * \brief Assigns additional data stamps to the registered global keyword
 *
 * The keyword supplied must already exist in the globals table, else an error occurs.
 * 
 * Returns 0 upon successful globals transmission, otherwise a TINE error code.
 *
 * @param keyword is the globals KEYWORD which identifies the data set being sent.
 * @param dataStamp is a user-defined integer value to accompany the globals multicast.
 * @param sysStamp is a system-defined integer value to accompany the globals multicast.
 *  
 */
TINE_EXPORT int AssignDataStampsToGlobal(char *keyword,int dataStamp,int sysStamp);

/**
 * \brief retrieves the link ID for a globals link
 *
 * The ID supplied as argument is the requested callback ID from the original link
 * registration.  This link ID is needed as argument to CloseLink().
 * 
 *
 * @param cbId is the callback ID passed as an argument to recvNetGlobalEx().
 *  
 * \return the link table ID for the globals link or a negative TINE error code.
 */
TINE_EXPORT int GetGlobalLinkIdFromCallbackId(int cbId);
#define getGlobalLinkIdFromCallbackId(a1) GetGlobalLinkIdFromCallbackId(a1)

/**
 * \brief retrieves the link ID for a globals link
 *
 * @param keyword is the globals keyword passed as an argument to recvNetGlobal().
 *  
 * \return the link table ID for the globals link or a negative TINE error code.
 */
TINE_EXPORT int GetGlobalLinkId(char *keyword);
#define getGlobalLinkId(a1) GetGlobalLinkId(a1)

/* SetGlobalsTableCapacity() */
/**
 * \brief sets the globals table capacity
 *
 * This establishes the maximum size of the (client-side) table which is to receive network globals.
 * (Default : 50).
 * 
 * @param value is the desired table size.
 *  
 * \note this call must be made prior to the client-side registration for any network global in order
 * to have any effect.
 *
 * \see GetGlobalsTableCapacity(), recvNetGlobal().
 */
TINE_EXPORT void SetGlobalsTableCapacity(int value);

/* GetGlobalsTableCapacity() */
/**
 * \brief gets the globals table capacity (client-side)
 *
 * \see SetGlobalsTableCapacity(), recvNetGlobal().
 */
TINE_EXPORT int GetGlobalsTableCapacity(void);

/* SetGCastTableCapacity() */
/**
 * \brief sets the globals multicast table capacity
 *
 * This establishes the maximum size of the (server-side) globals table which is to send network globals via
 * multicast.
 * (Default : 50).
 * 
 * @param value is the desired table size.
 *  
 * \note this call must be made prior to the server-side registration for any network global in order
 * to have any effect.
 *
 * \see GetGCastTableCapacity(), sendNetGlobal().
 */
TINE_EXPORT void SetGCastTableCapacity(int value);

/* GetGCastTableCapacity() */
/**
 * \brief gets the globals multicast table capacity (server-side)
 *
 * \see SetGCastTableCapacity(), recvNetGlobal().
 */
TINE_EXPORT int GetGCastTableCapacity(void);

/* SetGlobalsHeartbeat() */
/**
 * \brief sets the heartbeat (in seconds) during which time no 'timeout' messages will be sent 
 * to a globals dispatch routine (client-side)
 *
 * \param value is the desired heartbeat in seconds (default: 60 seconds)
 *
 * \see GetGlobalsHeartbeat(), recvNetGlobal().
 */
TINE_EXPORT void SetGlobalsHeartbeat(int value);

/* GetGlobalsHeartbeat() */
/**
 * \brief gets the heartbeat (in seconds) during which time no 'timeout' messages will be sent 
 * to a globals dispatch routine (client-side)
 *
 * \see SetGlobalsHeartbeat(), recvNetGlobal().
 */
TINE_EXPORT int GetGlobalsHeartbeat(void);

/**
 * \internal
 */
TINE_EXPORT int ListenForMulticasts(void);

/**
 * \internal
 */
TINE_EXPORT int ListenForNetGlobals(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* GLOBALS_H_ */
