#ifndef PROLOG_H_
#define PROLOG_H_

#include "project.def"
#include "tinedefs.h"
#include "errors.h"
#include "swaplib.h"
#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \brief Allocate memory for an array. Synonym for calloc() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param nr Number of elements.
 * \param sz Size of an element in bytes.
 */
#ifndef SystemCalloc
# define SystemCalloc(nr,sz) calloc(nr,sz)
#endif

/**
 * \brief Free allocated memory. Synonym for free() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param ptr pointer to memory area to be free()ed.
 */
#ifndef SystemFree
# define SystemFree(ptr) free(ptr)
#endif

#ifndef FEC_TIMEOUT
#define FEC_TIMEOUT      1   /* seconds */
#endif
#ifndef SERVER_TIMEOUT
#define SERVER_TIMEOUT   1   /* seconds */
#endif
#ifndef MAX_FDSETS
#define MAX_FDSETS      64
#endif
#ifndef PORTOFFSET
# define PORTOFFSET 0
#endif
#ifndef WATCHDOG
# define WATCHDOG 0
#endif

/* OS-specific definitions ... */
#if defined(MACOS)
#  define UNIX
#  define FREEBSD
#  define OS_TAG "MACOS"
#  define MIN_WORK_SIZE 0x10000
#endif
#if defined(DOS)              /* DOS */
#  define OS_TAG "DOS"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define SERVER_POLLING_MODE
#  define millisleep(ms) delay(ms)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#elif defined(UNIX)              /* UNIX */
#  ifndef OS_TAG
#    define OS_TAG "UNIX"
#  endif
#  define FS_DELIMITER '/'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  ifndef _REENTRANT
#    define _REENTRANT
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(char *)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#  define MIN_WORK_SIZE 0x10000
#elif defined(VXWORKS)         /* VxWorks */
#  define OS_TAG "VXWORKS"
#  define FS_DELIMITER '/'
#  define SCAN_FOR_NAME_LISTS FALSE
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(int)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#  define MIN_WORK_SIZE 2000
#  ifndef VXWTSK_FLAGS
#    define VXWTSK_FLAGS VX_FP_TASK
#  endif
#  ifndef VXWTSK_STACK
#    define VXWTSK_STACK 10000
#  endif
#  ifndef VXWTSK_PRIORITY
#    define VXWTSK_PRIORITY 100
#  endif
#elif defined(VMS)        /* VMS */
#  define OS_TAG "VMS"
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) 0
#  define unlink(nm) remove(nm)
#  define millisleep(ms) ;
#  define soclose(i) close(i)
#  define shutdown(s,t) close(s)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#elif defined(WIN32)         /* Win32 */
#  define OS_TAG "WIN32"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  define soerrno GetErrno()
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctlsocket(fd,REQ,(u_long *)buf)
#  define millisleep(ms) Sleep(ms)
#  define soclose(i) closesocket(i)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 0x10000
#elif defined(NIOS)          /* Altera NIOS */
#  define OS_TAG "NIOS"
#  define FS_DELIMITER '/'
#  define FS_RDONLY
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define ForceStructAlignment -1
#  define SERVER_POLLING_MODE
#  define millisleep(ms) sleep(ms)
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(u_long *)buf)
#  define soread(fd,buf,nbyte)  recv(fd,(char *)buf,nbyte,0)
#  define sowrite(fd,buf,nbyte) send(fd,(char *)buf,nbyte,0)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#else                       /* ????? */ 
#  ifndef OS_TAG
#    define OS_TAG "UNKNOWN"
#  endif
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#endif

#ifdef FS_RDONLY
# define NO_LTS
#endif

#ifndef SCAN_FOR_DEVICE_LISTS
# define SCAN_FOR_DEVICE_LISTS TRUE
#endif

#ifndef SCAN_FOR_NAME_LISTS
# define SCAN_FOR_NAME_LISTS TRUE
#endif

#ifndef ForceStructAlignment
# define ForceStructAlignment 0
#endif

#ifndef soerrno
#  define soerrno errno
#endif

#ifndef SERVER_POLLING_MODE
#  define SERVER_SELECT_MODE
#endif

/* 'us-or-them' definitions : */
#ifdef DOS   
#  include "dos.h"
#  define malloc_t void *
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT; }
#  define RDATE
#  define srvWorkArea     WorkArea
#  define srvWorkAreaSize WorkAreaSize
#  define clnWorkArea     WorkArea
#  define clnWorkAreaSize WorkAreaSize
#  define whoami(c) ;
#else
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT+size/5000; }
   void soperror(const char *txt);
#  ifndef soread
#    define soread(fd,buf,nbyte)  read(fd,(char *)buf,nbyte)
#  endif
#  ifndef sowrite
#    define sowrite(fd,buf,nbyte) write(fd,(char *)buf,nbyte)
#  endif
#  if !defined(SUNOS)
#    define malloc_t size_t
#  endif
#  if !defined(WIN32)
   char *strrev(char *s);
#  endif
#  if !defined(WIN32) && !defined(NIOS)
   int stricmp(char *s1, char *s2); /* from the DOS world */
   int strnicmp(char *s1, char *s2,int len); /* from the DOS world */
   char *strupr(char *s);
   char *strlwr(char *s);
#  endif
   extern BYTE *srvWorkArea;
   extern UINT32 srvWorkAreaSize;
   extern BYTE *clnWorkArea;
   extern UINT32 clnWorkAreaSize;
   char *whoami(char *user);
   void ClientDelay(int sec);  /* delay(0) never returns */
#endif                         /* DOS: definitions */

#ifndef VXWORKS
#  define inet_ntoa_b(n,b) strcpy(b,inet_ntoa(n))
#else
# if (_WRS_VXWORKS_MAJOR < 6)
  struct hostent
  {
    char *h_name;
    char **h_aliases;
    int  h_addrtype;
    int  h_length;
    char *h_addr;
  };
# endif
#endif

#ifndef UNIX
  TINE_EXPORT struct timeval *gettimeofday(struct timeval *t,struct timezone *tz);
#endif

#ifndef WIN32
  typedef int SOCKET;
#endif

#ifdef MULTITHREADED
# if defined(WIN32)
#   define WINTHREAD
# elif defined (UNIX)
#   define PTHREAD
# endif
# include "threader.h"
# define nDataOutBuffers 2
#else
# define nDataOutBuffers 1
#endif

void closeIPClients(void);

#ifndef IPXNW
typedef struct  /* an important entry from nwipxspx.h */
{
   BYTE network[4];
   BYTE node[6];
   BYTE socket[2];
} IPXAddress;
#endif

#define IsMemberIPControlNets(e,a) IsMemberIPNets((e)->numNetsAllowed,(e)->netsAllowed,a)
#define IsMemberIPBcastNets(a) IsMemberIPNets(NrOfIPBcastNets,IPBcastList,a)

#if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
# define doShmConsumer(n) DoClientEvents(n)
# define AcceptShm(n) DoServerEvents(n)
#else
# define doShmConsumer(n) ;
# define AcceptShm(n) ;
#endif
#if defined(IPXNW) || defined(IPXBSD)
#  define IsMemberIPXControlNets(addr) IsMemberIPXNets(NrOfIPXNets,IPXNetList,addr)
#  define IsMemberIPXBcastNets(addr) IsMemberIPXNets(NrOfIPXBcastNets,IPXBcastList,addr)
#  define sendGCast(buf,size)  sendIPgcast(buf,size); sendIPXgcast(buf,size)
#  define AcceptRequest(tv) { AcceptShm(100); AcceptIP(tv); AcceptIPX(tv); }
#  define Cleanup(tv)       { CleanupIPX(tv); CleanupIP(tv); CleanupStrandedDatagrams();}
#  define ntoa(c,b)     strcpy(b,\
   (*(UINT16 *)(c)->IPXaddress.socket ? ipxnet_ntoa((c)->IPXaddress.node) :\
                                        inet_ntoa((c)->IPaddress.sin_addr))\
   )
#  define tCall(i,buf,to) (ipLoaded ? tUDPCall(i,buf,to) : tIPXCall(i,buf,to))
#  define IsMemberControlNets(e,c) ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXControlNets(&(c)->IPXaddress) :\
				   IsMemberIPControlNets((e),&(c)->IPaddress)    )
#  define IsMemberBcastNets(c)   ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXBcastNets(&(c)->IPXaddress) :\
				   IsMemberIPBcastNets(&(c)->IPaddress)    )
   void closeIPXClients(void);
   int initIPXClient(int socket);
   int sendSPXData(WORD ID,int socket);
#  define ShutdownClients(i) {\
     closeIPXClients();\
     closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doShmConsumer(gBurstLimit);\
     doIPXConsumer();\
     doIPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#else
#  define sendGCast(buf,size) sendIPgcast(buf,size)
#  define AcceptRequest(tv) { AcceptShm(100); AcceptIP(tv); }
#  define Cleanup(tv)       { CleanupIP(tv); CleanupStrandedDatagrams(); }
#  define ntoa(c,b)         inet_ntoa_b((c)->IPaddress.sin_addr,b)
#  define tCall(i,buf,to)   tUDPCall(i,buf,to)
#  define IsMemberControlNets(e,c) IsMemberIPControlNets((e),&(c)->IPaddress)
#  define IsMemberBcastNets(c) IsMemberIPBcastNets(&(c)->IPaddress)
#  define sendSPXData(i,j) ipx_not_supported
#  define ShutdownClients(i) {\
	   closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doShmConsumer(gBurstLimit);\
     doIPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#endif         /* ! IPXNW || IPXBSD */

extern TINE_EXPORT int tineDebug;
extern int NGdebug;
int serverCycle(void);
int initTineResources(void);

#if defined(INTEL) || defined(VMS) || defined(NONALIGNED)
#  define PeekWord(w)    (w)
#  define PeekDWord(dw)  (dw)
#  define PrepWord(d,s)  { d = s; }
#  define PrepDWord(d,s) { d = s; }
#else
#  define PeekWord(w)    ((UINT16)ISWAP(w))
#  define PeekDWord(dw)  LSWAP(dw)
#  define PrepWord(d,s)  { d = (UINT16)ISWAP(s); }
#  define PrepDWord(d,s) { d = LSWAP(s); }
#endif

#ifdef DOS  /* user-defined structure swapping */
# define PrepOutgoingCLNQS(buf,c) memcpy(buf,c,sizeof(CLNQS));
#else
# define PrepOutgoingCLNQS(buf,c) {\
   _StructSwap(ClientQueryStructStruct,(BYTE *)buf,0,NULL,(BYTE *)c,LTOR);\
  }
#endif /* DOS */

#ifdef INTEL  /* Data Preparation MACROS: */

# define PrepSubData(pD,sz,f,dir) ;

# define PrepIncomingHeader(h,dat) { h = (PktHdr *)dat; }
# define PrepOutgoingSrvSettings(s,buf) {if ((void *)buf != (void *)s) memcpy(buf,s,SERVERSETTINGSSTRUCT_SIZE); }
# define PrepOutgoingAMS(amsstruct,buf,n) {if ((void *)buf != (void *)amsstruct) memcpy(buf,amsstruct,n*AMS_SIZE); }
# define PrepOutgoingAQS(a,buf) {if ((void *)buf != (void *)a) memcpy(buf,a,ACTIVITYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepOutgoingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPRPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PRPQUERYSTRUCT_SIZE); }
# define PrepOutgoingCONQS(c) ;
# define PrepOutgoingConTblInfo(p,buf,n)  {if ((void *)buf != (void *)p) memcpy(buf,p,n*CONTBL_QUERY_SIZE); };
# define PrepIncomingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepIncomingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPRPPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PRPQUERYSTRUCT_SIZE); }
# define PrepIncomingHRS(rb,buf) memcpy(rb,buf,HRS_SIZE);

# define AssignOutgoingHeader(h,buf) {\
    if (!ForceStructAlignment) h = (PktHdr *)buf;\
    memset(h,0,sizeof(PktHdr));\
  }
# define PrepOutgoingHeader(h,dat) { if (ForceStructAlignment) memcpy(dat,h,PKTHDR_SIZE); }

#else     /* non-INTEL world must SWAP ! */

# define PrepSubData(pD,sz,f,dir) {\
    SwapBytes((char *)pD,sz,LFMT(f),dir);\
}
# define PrepIncomingHeader(h,dat) {\
    _StructSwap(PktHdrStructStruct,(unsigned char *)dat,0,NULL,(unsigned char *)h,RTOL);\
}
# define PrepOutgoingSrvSettings(s,buf) {\
    _StructSwap(ServerSettingsStructStruct,(unsigned char *)buf,0,NULL,(unsigned char *)s,LTOR);\
}
# define PrepOutgoingAMS(amsstruct,buf,n)  {\
    _StructSwapEx(AMSStructStruct,(unsigned char *)buf,(unsigned char *)amsstruct,LTOR,n,sizeof(AMS));\
}
# define PrepOutgoingAQS(a,buf) {\
    _StructSwap(ActivityQueryStructStruct,(unsigned char *)buf,0,NULL,(unsigned char *)a,LTOR);\
}
# define PrepOutgoingPQS(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStruct));\
}
# define PrepOutgoingPQSX(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStructEx));\
}
# define PrepOutgoingXPQS(p,buf,n) {\
    _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(XPropertyQueryStruct));\
}
# define PrepOutgoingPRPQS(p,buf,n) {\
    _StructSwapEx(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PrpQueryStruct));\
}
# define PrepOutgoingCONQS(c) {\
    _LSWAP(c.contract.EqmSizeIn);\
    _LSWAP(c.contract.EqmSizeOut);\
    _LSWAP(c.pollingInterval);\
    _LSWAP(c.nclients);\
  }
# define PrepOutgoingConTblInfo(p,buf,n) {\
    _StructSwapEx(ConTblInfoStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,CONTBL_QUERY_SIZE);\
}
# define PrepIncomingPQS(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStruct));\
  }
# define PrepIncomingPQSX(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStructEx));\
  }
# define PrepIncomingXPQS(p,buf,n) {\
    _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(XPropertyQueryStruct));\
  }
# define PrepIncomingPRPQS(p,buf,n) {\
    _StructSwapEx(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PrpQueryStruct));\
  }
# define PrepIncomingHRS(rb,buf) {\
    _StructSwapEx(HRSStructStruct,(unsigned char *)buf,(unsigned char *)rb,RTOL,1,sizeof(HRS));\
  }
# define AssignOutgoingHeader(h,buf) memset(h,0,sizeof(PktHdr))
# define PrepOutgoingHeader(h,dat) {\
    _StructSwap(PktHdrStructStruct,(unsigned char *)dat,0,NULL,(unsigned char *)h,LTOR);\
  }

#endif  /* INTEL (data prepartion MACROS) */

/* client-side macros : */
#define IncrementConnections(n) (n)++

#define WaitForSyncCount(i) clientCycle()

#define CompIncomingData(out,cmp,sz,f,t) (\
     cmp == NULL ? FALSE :\
     t   != NULL ? !_datcmp(out,cmp,sz,f,t) :\
     !memcmp(out,cmp,sz)\
  )

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* PROLOG_H_ */
