
#ifndef SWAPLIB_H_
#define SWAPLIB_H_
/**
 * \internal
 * \file swaplib.h
 * \brief Header for swaplib.c
 * \see swaplib.c
 */
#include "project.def"

#include "tinedefs.h"
#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \internal
 * \brief Returns the address offset of a field in a structure.
 * 
 * \param struc  Structure name
 * \param field  Field name
 *
 */
#define OFFSETIN(struc,field)   ((size_t)&(((struc *)0L)->field))

#define MEMREV(d,s)\
  for (p1=(BYTE *)(d), p2=p1+(s)-1; p2 > p1; ++p1, --p2) {\
    *p1 ^= *p2; *p2 ^= *p1; *p1 ^= *p2;\
  }

#if defined(NONALIGNED) || defined(INTEL) || defined(VMS_IEEE)
#  define SwapBytes(dat,size,fmt,dir) ;
#  define ISWAP(w) (w)
#  define _ISWAP(w) (w)
#  define LSWAP(w) (w)
#  define _LSWAP(w) (w)
#  define _DSWAP(w) (w)
#else
#  ifdef VMS   /* ints unswapped, special swapping for floats, doubles */
#    define ISWAP(w) (w)
#    define _ISWAP(w) (w)
#    define LSWAP(w) (w)
#    define _LSWAP(w) (w)
#  else
#    define ISWAP(w) ((((w)&0xff00)>>8)|(((w)&0xff)<<8))
#    define _ISWAP(w) MEMREV(&(w),2)
#    define LSWAP(w) ((w>>24)|(w<<24)|((w&0xff00)<<8)|((w&0x00ff0000L)>>8))
#    define _LSWAP(w) MEMREV(&(w),4)
#  endif
#  define _DSWAP(w) MEMREV(&(w),8)
int SwapBytes(char *dat,int size,int fmt,int dir);
#endif
int cfHistorySwap(char *tag,BYTE *src,BYTE *dst,int dir,int num);

/**
 * \internal
 * \brief Format description of a structure element
 * 
 * To describe a whole structure, structFormat is used as
 * a linked list. In TINE, linked lists are mostly used in
 * reverse order.
 */
struct structFormat
{
  int siz; /**< number of elements. */
  int fmt; /**< TINE CF_ format. */
  int off; /**< offset in the byte stream */
  int addr; /**< Byte offset within the structure. */
  char field[FIELD_NAME_SIZE]; /**< structure field name */
  char stag[TAG_NAME_SIZE]; /**< sub tag if fmt == CF_STRUCT */
  struct structStruct *owner; /**< reference to struct owner of this structFormat */
  struct structFormat *nxt; /**< Pointer to next element. */
};

/**
 * \internal
 * \brief Structure description for tagged structures
 * 
 * A structStruct contains a structFormat list of fields. 
 * It does not only describe a structure, but also describes
 * an array of data of that structure. 
 * At the same time, it is also a buffer.
 * It can also be used as an element of a linked list.
 * 
 * A structStruct is used for tagged structures.
 */
struct structStruct
{
  char name[TAG_NAME_SIZE];/**< structure tag */
  int  rawlen;             /**< raw/network data size of the described structure */
  int  size;               /**< local structure size (memory size) of the described structure */
  int  num;                /**< maximum number of structure sets buffered (array size) */
  int  sealed;             /**< structure has been sealed */
  int  hasExtSpace;        /**< has extended space (uses CF_STRING, CF_IMAGE, etc.) */
  BYTE *sdata;             /**< buffer for data to be sent over the network */
  struct structFormat *s;  /**< structure definition */
  struct structStruct *next;/**< pointer to the next element in a linked list.*/
};

/**
 * \internal
 * \brief Global list of tagged structures.
 */
extern struct structStruct *structList;

struct structStruct *getStructFromTag(char *tag);
struct structStruct *findstruct(char *tag);
int addtostruct(void *target,int addr,int size, int fmt, char *subtag, char *field);
int getFieldFormat(struct structStruct *s,char *field);
TINE_EXPORT int fillstruct(char *tag,int addr,int size,int fmt);

/**
 * \brief Adds a field description to a tagged structure.
 * 
 * Use this routine to register the fields of your structure within
 * the structure registry.  The fields can be any TINE format type
 * or another tagger structure (in which case fmt = CF_STRUCT, and
 * field = <parent tag>fieldname).
 *
 * \param tag structure tag name
 * \param addr is the address offset of this field inside the structure.
 * \param size is the array length if the described field. (not the size in bytes)
 * \param fmt format of the new struct element
 * \param field field name
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see SealTaggedStruct
 * 
 * \b Example:
 *
 * \include eg_taggedStruct3.c
 * 
 */
TINE_EXPORT int AddFieldToStruct(char *tag,int addr,int size,int fmt,char *field);
#define addFieldToStruct(a1,a2,a3,a4,a5) AddFieldToStruct(a1,a2,a3,a4,a5)

/**
 * \brief Returns the 'raw' structure length of the tagged structure given.
 * 
 * This routine will retrieve the raw structure length of the tagged structure
 * given from the structure registry.  This length is the length in bytes of the
 * pure byte stream and is NOT necessary identical to the 'sizeof()' call.
 * These two values will only be the same if the registered structure is properly 
 * aligned.
 *
 * \param tag structure tag name
 * 
 * \return the size in bytes of the raw structure length or the negative of a 
 * TINE return code.
 * 
 */
TINE_EXPORT int GetStructSize(char *tag);

/**
 * \brief Seals a tagged structure (registration finished!).
 * 
 * Call this routine to signal the structure registry that there are
 * no more fields to be added to the tagged structure.
 * 
 * \param tag structure tag name
 * \param size total size of the structure in bytes
 * \param number number of elements of this structure to allocate
 * 
 * \return 0, un_allocated, out_of_local_memory 
 * 
 * \include eg_taggedStruct.c
 * 
 * \see AddFieldToStruct
 */ 
TINE_EXPORT int SealTaggedStruct(char *tag,int size,int number);
#define sealTaggedStruct(a1,a2,a3) SealTaggedStruct(a1,a2,a3)
int _StructSwap(struct structFormat *s,BYTE *dat,int offset,int *tail,BYTE *struc,int dir);
int StructSwap(char *tag,BYTE *dat,BYTE *struc,int dir);
int StructSwapEx(char *tag,BYTE *dat,BYTE *struc,int dir,int num,int structsize);
int _StructSwapEx(struct structFormat *s,BYTE *dat,BYTE *struc,int dir,int num,int structsize);
int initControlStructs(void);
void RegisterLongDataStructs(void);
int queryStructFormat(char *tag,BYTE *sfq,int fmt,int *sfqsiz);
void freeStruct(struct structFormat *sf);
void freeStructList(int freeMarkedList);
struct structStruct *getTaggedStruct(char *tag);
TINE_EXPORT void flushStructCache(void);

/**
 * \brief Field list of structure PktHdr. 
 */
extern struct structFormat *PktHdrStructStruct;

/**
 * \brief Field list of structure CONTRACT.
 */
extern struct structFormat *CONTRACTStruct;
extern struct structFormat *CONTRACT5Struct;

/**
 * \brief Field list of structure ProducerHeaderStruct.
 */
extern struct structFormat *PrdrHdrStruct;
extern struct structFormat *PrdrHdr5Struct;

/**
 * \brief Field list of structure SubInfoPkt.
 */
extern struct structFormat *SubInfoPktStruct;

/**
 * \brief Field list of structure SubInfoPkt.
 */
extern struct structFormat *SubInfoPkt5Struct;

/**
 * \brief Field list of structure RPCDataStruct.
 */
extern struct structFormat *RPCDataStructStruct;

/**
 * \brief Field list of structure RPCFecStruct.
 */
extern struct structFormat *RPCFecStructStruct;

/**
 * \brief Field list of structure ADSStruct.
 */
extern struct structFormat *ADSStructStruct;
extern struct structFormat *ADS5StructStruct;

/**
 * \brief Field list of structure AMSStruct.
 */
extern struct structFormat *AMSStructStruct;
extern struct structFormat *AMS5StructStruct;

/**
 * \brief Field list of structure AWSStruct.
 */
extern struct structFormat *AWSStructStruct;

/**
 * \brief Field list of structure PropertyQueryStruct.
 */
extern struct structFormat *PropertyQueryStructStruct;

/**
 * \brief Field list of structure PropertyQueryStructEx.
 */
extern struct structFormat *PropertyQueryStructExStruct;

/**
 * \brief Field list of structure XPropertyQueryStruct.
 */
extern struct structFormat *XPropertyQueryStructStruct;

/**
 * \brief Field list of structure PrpQueryStruct.
 */
extern struct structFormat *PrpQueryStructStruct;

/**
 * \brief Field list of structure ActivityQueryStruct.
 */
extern struct structFormat *ActivityQueryStructStruct;

/**
 * \brief Field list of structure ServerStateStruct.
 */
extern struct structFormat *ServerStateStructStruct;

/**
 * \brief Field list of structure GlbDataHdr3.
 */
extern struct structFormat *GlbDataHdr3Struct;

/**
 * \brief Field list of structure GlbDataHdr.
 */
extern struct structFormat *GlbDataHdrStruct;

/**
 * \brief Field list of structure ClientQueryStruct.
 */
extern struct structFormat *ClientQueryStructStruct;

/**
 * \brief Field list of structure hdrSPECTRUM.
 */
extern struct structFormat *SpectrumHeaderStruct;

/**
 * \brief Field list of structure hdrIMAGE.
 */
extern struct structFormat *ImageHeaderStruct;

/**
 * \brief Field list of structure HRSstruct.
 */
extern struct structFormat *HRSStructStruct;

extern struct structFormat *FDSStructStruct;

extern struct structFormat *ConTblInfoStructStruct;

extern struct structFormat *WriteAccessQueryStruct;

extern struct structFormat *ServerSettingsStructStruct;

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* SWAPLIB_H_ */
